/*#########################################################################
  # Programm # pendel.c                                                    #
  #          #                                                             #
  # Compiler # Turbo Assembler 3.0 / Borland C++ 3.1                       #
  #          #                                                             #
  # Datum    # Uni Karlsruhe, 13.01.97                                     #
  #          #                                                             #
  # Autor    # Frank Ruhnau                                                #
  #########################################################################
  # Inhalt   # Pendelversuch, Simulation und Messung in Echtzeit,          #
  #          # Interruptgesteuerte Datenerfassung bzw. -verarbeitung       #
  #          #                                                             #
  #          #                                                             #
  ##########################################################################*/
/*##########################################################################
  # Programmkopf                                                           #
  ##########################################################################*/
/*##########################################################################
  # Includedateien einladen                                                #
  ##########################################################################*/
#include <graphics.h>
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include <math.h>
#include <time.h>
#include <string.h>
/*##########################################################################
  # Konstanten- und Variablendeklaration                                   #
  ##########################################################################*/
/*##########################################################################
  #  Konstanten fuer Tastaturabfrage                                       #
  ##########################################################################*/
#define F1   59          //F1-Taste
#define F2   60          //F2-Taste
#define F3   61          //F3-Taste
#define F4   62          //F4-Taste
#define F5   63          //F5-Taste
#define F6   64          //F6-Taste
#define F7   65          //F7-Taste
#define F8   66          //F8-Taste
#define F9   67          //F9-Taste
#define F10  68          //F10-Taste
#define F11  133         //F11-Taste
#define F12  134         //F12-Taste
#define ESC  27    //Escape-Taste
/*##########################################################################
  #  Konstanten fuer den Interruptcontroller 8259A                         #
  ##########################################################################*/
#define eoi 0x20         //End of Interrupt, Befelscode
#define irq12   0x74       //IRQ12 = Vektor
#define master 0x20      //Adr Kontrollreg. des 1.Int.controllers
#define mastermask 0x21  //Adr Kontrollreg. des 1.Int.controllers (Maske)
#define slave 0x0A0      //Adr Kontrollreg. des 2.Int.controllers
#define slavemask 0x0A1  //Adr Kontrollreg. des 2.Int.controllers (Maske)
/*##########################################################################
  #  Konstanten fuer den Timerbaustein 8253                                #
  ##########################################################################*/
#define timer1 0x305     //Adresse Datenregister Zaehler 1
#define timercontr 0x307 //Adresse Controlregister
#define t1mode  0x76       //Timer1, Mode 3,16 Bit (symmetr. Teiler)
/*##########################################################################
  #  Konstanten fuer den I/O-Baustein 8255                                 #
  ##########################################################################*/
#define IO_PORTC 0x302   //Adresse Port C
#define IO_CONTR 0x303   //Adresse Controllregister
#define IO_INIT  0x9A      //nur PortC Bit 0-3 als Ausgang 0x9A
#define IO_RESET 0x9B      //alle Ports auf Eingang
/*##########################################################################
  #  Konstanten fuer AD-Wandlung                                           #
  ##########################################################################*/
#define EPLD        0x308    //Adresse Datenregister Zaehler 1
#define AUFLOESUNG  0        //0 = 12 Bit, 8 = 8 Bit
#define OFFSET      2047.5   //Offset AD-Wandler bei Winkel 0
#define MAXWINKEL   M_PI/3.0 //max. Winkelauslenkung
#define WINKELAUFL  MAXWINKEL/4095.0  //Winkelaufloesung bei 12Bit
#define MAXOMEGA    2.0      //max. Winkelgeschw. fuer Phasenportraet
#define MAXTIME     340      //max. Zeit fuer Amplitudendarstellung
/*##########################################################################
  # Konstanten fr Grafik                                                  #
  ##########################################################################*/
#define S_XMODEL       75    //Drehpunkt der Pendelachse fuer Simulation
#define S_YMODEL       15

#define M_XMODEL       395   //Drehpunkt der Pendelachse fuer Messung
#define M_YMODEL       15

#define M_XPORT          533   //Koordinatenursprung/Phasenportraet/Messung
#define M_YPORT        113

#define S_XPORT          213   //Koordinatenursprung/Phasenportraet/Simulation
#define S_YPORT        113

#define M_XPORT_SKAL   80    //Skalierung/Phasenportraet/Messung
#define M_YPORT_SKAL   36

#define S_XPORT_SKAL   80    //Skalierung/Phasenportraet/Simulation
#define S_YPORT_SKAL   36

#define M_XAMPLI       340   //Koordinatenursprung/Amplitudenverlauf/Mess.
#define M_YAMPLI       295

#define S_XAMPLI       20    //Koordinatenursprung/Amplitudenverlauf/Simu.
#define S_YAMPLI       295

#define M_YAMPLI_SKAL  90    //Skalierung/Amplitudenverlauf/Messung
#define S_YAMPLI_SKAL  90    //Skalierung/Amplitudenverlauf/Simulation
#define XAMPLI_SKAL    0.8   //Skalierung/Amplitudenverlauf/Simu.+ Mess.

#define L_MODEL_SKAL   80    //Pendellaenge fuer Simmulation und Messung
/*##########################################################################
  #  sonstige Konstanten                                                   #
  ##########################################################################*/
/*##########################################################################
  # Globale Variablen                                                      #
  ##########################################################################*/
int irq12status = 0;        //Interruptstatusflag fr IRQ 12
unsigned int error_flag=0;  //Error-Flag ungleich 0 ->Fehler,IRQ ausgelassen
unsigned int errors=0;      //Error-Counter
unsigned int irq_run = 0;   //zum aktivieren der Intrruptseviceroutine

float irqfreq_in = 100.0;   //Min. 31 (wegen Timer), Max. 25000 (wegen ADW)
float irqfreq;        //realisierbare Frequenz von irqfreq_in
unsigned int t1_teiler;     //Timer1 Teilerfaktor

void interrupt (*old_vect12) (void);//Zeiger auf alte Vektorroutine

double dt;                  //Zeitschritt fuer numerische Integration
/*##########################################################################
  # Globale Schalter-Variablen fuer Menue                                  #
  ##########################################################################*/
char s_model=1, s_portraet=1, s_ampli=1, s_synchron=0, s_peri=1;
char m_model=1, m_portraet=1, m_ampli=1,               m_peri=1;
/*##########################################################################
  # Globale Variablen Menue-Strings                                         #
  ##########################################################################*/
char text[100], textin[100];
/*##########################################################################
  # Globale Variablen fuer Simulation                                      #
  ##########################################################################*/
const double teta =0.623 ;             //gesamte Traegheitsmasse
const double s= 0.33432;                //Schwerpunktsabstand der Gesamtmasse
const double m= 4.60755;                //Gesamtmasse
const double g= 9.80665;              //Erdbeschleunigung
double r      = 0.2;                      //Reibung
double d      = 0;                      //Pendeldaempfung
double s_p    = M_PI/180.0*52.0;      //Anfangswert Winkelauslenkung
double s_vp   = 0;                      //           Winkelgeschwindigkeit
double c_p, c_vp;     //Koefizienten von Winkel u. Winkelgeschwindigkeit
int ommmm = 100;

int    s_xpix=S_XPORT, s_ypix=S_YPORT;
float  s_t=0;           //Laufvariable fuer Zeitachse Amplitudenverlauf
unsigned int s_x=S_XMODEL, s_y=S_YMODEL;//Ursprungskoordinaten,Pendelmodell
/*##########################################################################
  # Globale Variablen fuer Messung                                         #
  ##########################################################################*/
unsigned int ad_wert=0  ;                   //eigelesener Wert von ADW
unsigned int ad_wert_alt=0xffff;
double m_p;                                 //Winkelauslenkung
unsigned int m_max;                       //Maxwert der Winkelauslenkung
int    m_xpix=M_XPORT, m_ypix=M_YPORT;
float  m_t=0;                               //Amplitudenverlauf
unsigned int m_x=M_XMODEL, m_y=M_YMODEL;  //Koordinaten Pendelmodel
/*##########################################################################
  # Diese Includedateien werden erst nach der Definition der darin         #
  # benoetigten Variablen eingebunden                                      #
  ##########################################################################*/
#include "pen_graf.h"           //Bild- und Menueaufbaust
#include "pen_hw.h"           //Initialisierung der PC-Hardware
/*##########################################################################
  # Prototypendeklaraton                                                   #
  ##########################################################################*/
void graphinit(void);                   //in "pen_hw.h" definiert
void initinter(void);                   //in "pen_hw.h" definiert
void initports(void);                 //in "pen_hw.h" definiert
void inittimer(void);                 //in "pen_hw.h" definiert
void interrupt read_ad(void);         //Interruptroutine fuer Motorsteuerung
void resetinter (void);   //          //in "pen_hw.h" definiert
unsigned long  runden(float);
void runge_kutta(void);
void mes_winkel(void);
void mes_model(void);
void mes_portraet(void);
void mes_ampli(void);
void sim_model();
void sim_portraet(void);
void sim_ampli(void);
void sim_synchron(void);
void menue_tasten(int);         //in "pen_graf.h" definiert
void menue(void);                       //in "pen_graf.h" definiert
void bildaufbau(void);          //in "pen_graf.h" definiert
void init_param(void);
void irq_freq(void);
void sim_peri(void);
void mes_peri(void);

#define TPOSX 8
#define TPOSY 160
#define MPOSX 328
#define MPOSY 160


/*##########################################################################
  # Unterprogramme                                                         #
  ##########################################################################*/
/*##########################################################################
  # Koefizientenberechnung fuer Simulation                                 #
  ##########################################################################*/
void init_param(void)
{
	double dummy;
	c_p       = -m*g*s/teta ;          //Koeffizient fuer Winkel
	c_vp      = -d*s/teta ;            //Koeffizient fuer Winkelgeschw.
	dummy     = 2000000.0/irqfreq_in;  //Teilerwert fuer Timer1 berechnen
	dummy     = runden(dummy);
	t1_teiler = (unsigned int)dummy;   //Teilerwert fuer Timer1
	irqfreq   = 2000000.0/dummy;       //realisierbare Frequenz berechnen
	dt        = 1.0/irqfreq;           //Zeitschritt numerische Integration
}
/*##########################################################################
  #  Interruptroutine ADW lesen                                            #
  ##########################################################################*/
void interrupt read_ad(void)
{
	disable();        //Interrupts abschalten whrend der Routine
			  //vorheriger IRQ12  verarbeitet?
	if ( irq12status == 0 && irq_run == 1) {
		irq12status = 1;         //neuer IRQ12 anzeigen
		ad_wert = inport(EPLD);  //gewandelter Wert
	} else if ( irq12status == 1 && irq_run == 1) {
		error_flag = 1;
	}
	outportb(master, eoi);      //End-Of-Interrupt and Interr.-Controller 1
	outportb(slave, eoi);       //End-Of-Interrupt and Interr.-Controller 2
	enable();                   //Interrupts allgemein anschalten
}
/*##########################################################################
  # Kommazahl nach Ganzzahl runden                                         #
  ##########################################################################*/
unsigned long runden(float a)
{
	unsigned long b, c;
	b = a;                   //b enthlt Ganzzahl von a
	c = ((a-b) > 0.5) ? b+1 : b;
	return(c);               //gerundete Zahl zurckgeben
}
/*##########################################################################
  # Periodendauermessung bei Simulation                                    #
  ##########################################################################*/
void sim_peri(void)
{
	static float s_p_old = 0;
	static float schwingungs_zeit = 0;

	schwingungs_zeit += dt;

	if (s_p_old < 0 && s_p > 0) {
		sprintf(text, "T=%f s", schwingungs_zeit);
		bar(TPOSX,TPOSY,TPOSX+10*8,TPOSY+8);
		outtextxy(TPOSX,TPOSY,text);
		schwingungs_zeit = 0;
	}
	s_p_old = s_p;
}
/*##########################################################################
  # Periodendauermessung bei Messung                                       #
  ##########################################################################*/
void mes_peri(void)
{
	static float schwingungs_zeit = 0;
	static float m_p_old = 0;
	static char schalter = 0;

	schwingungs_zeit += dt;

	if (m_p_old < 0 && m_p > 0) {
		sprintf(text, "T=%f s", schwingungs_zeit);
		bar(MPOSX,MPOSY,MPOSX+10*8,MPOSY+8);
		outtextxy(MPOSX,MPOSY,text);
		schwingungs_zeit = 0;
	}
	if (m_p > m_p_old) schalter = 1;
	if (m_p < m_p_old && schalter) {
		schalter = 0;
		ommmm++;
	}
	m_p_old = m_p;
}
/*##########################################################################
  # Numerische Integration nach Runge-Kutta                                #
  ##########################################################################*/
void runge_kutta(void)
{
	float k[4];
	float m[4];
	float K,M;
	float s_p_old, s_vp_old;
	float sgn_s_vp;
	int i;

	s_p_old = s_p;
	s_vp_old = s_vp;

	for (i = 0; i < 3; i++) {
		sgn_s_vp = (s_vp > 0) ? 1 : (s_vp < 0) ? -1 : 0;
		k[i] = dt * s_vp;
		m[i] = dt * (d * s / teta * s_vp + c_p * sin(s_p) - r * sgn_s_vp);
		s_p = s_p_old + k[i] / 2;
		s_vp = s_vp_old + m[i] / 2;
	}

	s_p += k[2] / 2;
	s_vp += k[2] / 2;

	sgn_s_vp = (s_vp > 0) ? 1 : (s_vp < 0) ? -1 : 0;
	k[3] = dt * s_vp;
	m[3] = dt * (d * s / teta * s_vp + c_p * sin(s_p) - r * sgn_s_vp);

	K = (k[0] + 2 * k[1] + 2 * k[2] + k[3]) / 6;
	M = (m[0] + 2 * m[1] + 2 * m[2] + m[3]) / 6;

	s_p = s_p_old + K;
	s_vp = s_vp_old + M;
}
/*##########################################################################
  # Winkelberechnung fuer Messung                                          #
  ##########################################################################*/
void mes_winkel(void)
{
	m_p=(ad_wert-OFFSET)/OFFSET*MAXWINKEL;
}
/*##########################################################################
  # Synchronisierung der Simulation zur Messung                            #
  ##########################################################################*/
void sim_synchron(void)
{
	if (ommmm > 1 || ommmm == 0) {
	    ommmm = 0;
	    return;
	}
	s_p = m_p;
	s_vp = 0;
	s_synchron = 0;
}
/*##########################################################################
  # Pendel-Model fuer Simulation                                           #
  ##########################################################################*/
void sim_model(void)
{
	line(S_XMODEL, S_YMODEL, s_x, s_y);           //lscht alte Linie
	s_x     = runden(S_XMODEL + sin(s_p)*L_MODEL_SKAL);
	s_y     = runden(S_YMODEL + cos(s_p)*L_MODEL_SKAL);
	line(S_XMODEL, S_YMODEL, s_x, s_y);
}
/*##########################################################################
  # Pendel-Model fuer Messung                                              #
  ##########################################################################*/
void mes_model(void)
{
	line(M_XMODEL, M_YMODEL, m_x, m_y);           //loescht alte Linie
	m_x     = runden(M_XMODEL + sin(m_p)*L_MODEL_SKAL);
	m_y     = runden(M_YMODEL + cos(m_p)*L_MODEL_SKAL);
	line(M_XMODEL, M_YMODEL, m_x, m_y);
}
/*##########################################################################
  # Phasenportraet fuer Simulation                                         #
  ##########################################################################*/
void sim_portraet(void)
{
	float y, x;
	line(S_XPORT, S_YPORT, s_xpix, s_ypix);          //alte Linie loeschen
	y = S_YPORT - S_YPORT_SKAL * s_vp;
	x = S_XPORT + S_XPORT_SKAL * s_p;
	s_ypix = (int) runden(y);
	s_xpix = (int) runden(x);
	putpixel(s_xpix, s_ypix, RED);
	line(S_XPORT, S_YPORT, s_xpix, s_ypix);
}
/*##########################################################################
  # Phasenportraet fuer Messung                                            #
  ##########################################################################*/
void mes_portraet(void)
{
	float d_phi;
	float w, y, x;
	line(M_XPORT,M_YPORT,m_xpix, m_ypix);          //alte Linie loeschen
	d_phi = WINKELAUFL*(int)(ad_wert-ad_wert_alt); //Delta Winkel
	w     = d_phi * irqfreq;                    //Winkel-Geschwindigkeit
	y     = M_YPORT - w * M_YPORT_SKAL;
	m_ypix = (int)runden(y);
	x = M_XPORT + m_p * M_XPORT_SKAL;
	m_xpix = (int)runden(x);
	putpixel(m_xpix, m_ypix, GREEN);
	line(M_XPORT, M_YPORT, m_xpix, m_ypix);
}
/*##########################################################################
  # Amplituten-Zeit-Diagramm fuer Simulation                               #
  ##########################################################################*/
void sim_ampli(void)
{
	s_t += dt*XAMPLI_SKAL;
	putpixel(S_XAMPLI+s_t, S_YAMPLI + s_p*S_YAMPLI_SKAL, YELLOW);
}
/*##########################################################################
  # Amplituten-Zeit-Diagramm fuer Messung                                  #
  ##########################################################################*/
void mes_ampli(void)
{
	m_t += dt*XAMPLI_SKAL;
	putpixel(M_XAMPLI+m_t, M_YAMPLI + m_p*M_YAMPLI_SKAL, YELLOW);
}
/*###########################################################################
  # Eingabemenue fuer IRQ-Frequenz                                          #
  ##########################################################################*/
void irq_freq(void)
{
	setfillstyle(SOLID_FILL,YELLOW);
	bar(185,345,318,392);           //Textfeld anzeigen
	gotoxy(25,23);
	printf("neue IRQ-Freq.:\n\t\t\t");
	gets(textin);
	sscanf(textin,"%f",&irqfreq_in);
	setfillstyle(SOLID_FILL,BLACK);
	bar(185,345,318,392);           //Textfeld loeschen
	init_param();
	inittimer();
}
/*##########################################################################
  # Hauptprogramm                                                          #
  ##########################################################################*/
int main(void)
{  
	int taste, ende=0;
	clrscr();
	graphinit();
	setwritemode(XOR_PUT);
	setfillstyle(SOLID_FILL,BLACK);      //um Grapfikbereiche zu loeschen
	initports();
	initinter();
	init_param();
	inittimer();
	bildaufbau();
	menue();
	irq_run=1;    //IRQ-Serviceroutine wird durchlaufen wenn 1
	do {
		//  irq12status=1;       //hilfreich beim Steppen
		//  delay(5);            //nur hilfreich ohne IRQs
		if(irq12status==1) {     //IRQ12 bereits eingetroffen
			irq12status = 0;        //IRQ12 rcksetzen
			runge_kutta();          //numerische Integration
			mes_winkel();           //Winkelberechnung aus ad_wert
			//Synchronisierung Simulation
			if(s_synchron == 1) sim_synchron(); 
			//grafisches Pendel/Simul.
			if(s_model    == 1) sim_model();  
			//grafisches Pendel/Messung
			if(m_model    == 1)  mes_model();   
			
			//Phasenportraet/Simul.
			if(s_portraet == 1)  sim_portraet(); 
			//Phasenportraet/Messung
			if(m_portraet == 1) {   
				if(ad_wert_alt!=0xffff) mes_portraet();
				ad_wert_alt=ad_wert;
			}

			//Periodendauermessung/Simul.
			if(s_peri     == 1)  sim_peri();   
			//Periodendauermessung/Messung
			if(m_peri     ==1)  mes_peri();     

			//Amplitudenverlauf/Siumul.
			if(s_ampli   ==1)  sim_ampli();   
			//Amplitudenverlauf/Messung
			if(m_ampli   ==1)  mes_ampli();   
		}
		if(error_flag == 1) {
			irq_run=0;
			errors++;       
			error_flag=0;
			menue();
			ad_wert_alt=0xffff;     //Initialisieren
			irq_run=1;
		}
		if (kbhit()) {
			irq_run=0;              //IRQ-Serviceroutine inaktiv
			taste=getch();
			if(!taste)         taste=getch();
			if(taste==ESC)     ende=1;        //exit(EXIT_SUCCESS);
			menue_tasten(taste);                //Hauptmenue
			taste=0;                //taste rcksetzen
			ad_wert_alt=0xffff;     //Initialisieren
			irq_run=1;              //IRQ-Serviceroutine aktiv
		}
	} while(ende!=1);
	resetinter();
	outportb(timercontr,t1mode); //Reset 8353, Zhler stop
	outportb(IO_CONTR,IO_RESET); //Reset 8255, alle Ports auf Eingang
	//     cleardevice();             //lscht Grafikbildschirm
	//     closegraph();              //schliet das Graphiksystem
	return(0);                   //Rueckgabewert an DOS
}
