/*ͻ
   Programm  Versuch0.c                                                  
                                                                         
   Compiler  Borland C++ 3.1                                             
                                                                         
   Datum     Karlsruhe, den 04.07.94                                     
                                                                         
   Update                                                                
                                                                         
   Autor     Hartung Wilstermann & Ingo Gabriel                          
  Ķ
   Inhalt    Schrittmotorsteuerung                                       
  ͼ*/

/*͸
   Includedateien                                                         
  ;*/
  #include <graphics.h>         //Graphicmodul
  #include <stdlib.h>           //Standardbibliothek
  #include <stdio.h>            //Ein-/Ausgabe
  #include <conio.h>            //Bildschirm / Tastatur
  #include <dos.h>              //DOS- und Interruptfunktionen
  #include <math.h>             //Mathematische Funktionen
  #include <string.h>           //Stringbehandlung

/*͸
   Konstanten- und Variablendeklaration                                   
  ;*/
/*Ŀ
    Konstanten fuer Tastaturabfrage                                       
  */
  #define F1 59                 //F1-Taste
  #define F2 60                 //F2-Taste
  #define F3 61                 //F3-Taste
  #define F4 62                 //F4-Taste
  #define F5 63                 //F5-Taste
  #define F6 64                 //F6-Taste
  #define F7 65                 //F7-Taste
  #define F8 66                 //F8-Taste
  #define F9 67                 //F9-Taste
  #define F10 68                //F10-Taste
  #define CURSORRIGHT 77        //Cursor rechts
  #define CURSORLEFT 75         //Cursor links
  #define CURSORUP 72           //Cursor oben
  #define CURSORDOWN 80         //Cursor unten
  #define PAGEUP 73             //Bild nach oben
  #define PAGEDOWN 81           //Bild nach unten
  #define INS 82                //Einfgetaste
  #define DEL 83                //Entferntaste
  #define BACKSPACE 8           //Backspacetaste
  #define TAB 9                 //Tabulator
  #define LEER 32               //Leertaste


/*Ŀ
    Konstanten fuer den I/O-Baustein 8255                                 
  */
  #define porta 0x300         	//Adresse Port A
  #define portb 0x301         	//Adresse Port B
  #define portc 0x302         	//Adresse Port C
  #define cntlreg 0x303       	//Adresse Kontrollregister
  #define lslreset 0x9B       	//Steuerwort alle Ports auf Eingang
  #define setio 0x89          	//Port A,B Ausgang, Port C Eingang

/*Ŀ
    Konstanten fuer den Interruptcontroller 8259A                         
  */
  #define eoi 0x20              //End of Interrupt
  #define irq10	0x72            //IRQ10 = Vektor
  #define irq11	0x73            //IRQ11 = Vektor
  #define master 0x20           //Adr Kontrollreg. des 1.Int.controllers
  #define mastermask 0x21       //Adr Kontrollreg. des 1.Int.controllers (Maske)
  #define slave 0x0A0           //Adr Kontrollreg. des 2.Int.controllers
  #define slavemask 0x0A1       //Adr Kontrollreg. des 2.Int.controllers (Maske)

/*Ŀ
    Konstanten fuer den Timerbaustein 8253                                
  */
  #define timer0 0x304          //Adresse Datenregister Zaehler 0
  #define timer1 0x305          //Adresse Datenregister Zaehler 0
  #define timer2 0x306          //Adresse Datenregister Zaehler 0
  #define timercontr 0x307      //Adresse Controlregister

/*Ŀ
    globale Variable                                                      
  */
  int sinuswerte[32];           //Ergebnisse der Sinusberechnung
  int cosinuswerte[32];         //Ergebnisse der Cosinusberechnung
  int irqstatus;		            //Interruptstatusflag
  void interrupt (*old_vect10) (void);//Zeiger auf alte Vektorroutine
  void interrupt (*old_vect11) (void);//Zeiger auf alte Vektorroutine

/*Ŀ
    Functionsprototypen                                                   
  */
  void graphinit(void);		//Graphiktreiber und Tacho initialisieren
  void tacho(unsigned int dreh);//stellt Drehzahl graphisch dar
  void sinuscosinus (void);     //Sinus und Cosinuswerte berechnen
  void initports (void);        //Initialisierung des parallelen I/O-Bausteins 8255
  void inittimer (void);        //Timer vorbereiten
  void initinter (void);        //Einfuegen der Interruptroutine in die Vektortabelle und Int freigeben
  void interrupt interruptmotor(void); //Interruptroutine fuer Motor
  void interrupt interruptlight(void); //Interruptroutine fuer Lichtschranke
  void resetinter (void);       //Interrupts sperren und alte Interruptroutine wieder eintragen
  void resetports (void);       //alle Ports auf Eingang
  void loadram(void);           //RAM's laden

/*͸
    Unterprogramme                                                        
  ;*/
/*Ŀ
   Graphiktreiber initialisieren                                          
  */
  void	graphinit(void)
  {
    int gdriver = DETECT, gmode, errorcode;

    initgraph(&gdriver,&gmode,""); 	//Graphiktreiber und-mode ermitteln
    errorcode = graphresult();
    if (errorcode != grOk)
    {
      cprintf("Graphikfehler: %s\n",grapherrormsg(errorcode));
      cprintf("Drcken Sie eine Taste um anzuhalten:");
      getch();
      exit(1);
    };
    setbkcolor(BLACK);          //Hintergrundfarbe setzen
    init_tacho();               //tacho initialisieren
  };

/*Ŀ
   Berechnet Sinus- und Cosinusfunktion mit 8 Bit Breite (0..255)         
  */
  void sinuscosinus (void)
  {
    int i;

    for (i=0; i != 32; i++)
    {
      sinuswerte[i]=255/2.0+sin(i*M_PI*2/((float)32))*254/2.0;
      cosinuswerte[i]=255/2.0+cos(i*M_PI*2/((float)32))*254/2.0;
    }
  };


/*Ŀ
    Initialisierung des parallelen I/O-Bausteins 8255                     
  */
  void initports (void)
  {

  };


/*Ŀ
    Timer vorbereiten                                                     
  */
  void inittimer (void)
  {

  };



/*Ŀ
    Einfuegen der Interruptroutine in die Vektortabelle und Int freigeben 
  */
  void initinter (void)
  {
     old_vect10= getvect(irq10); //alte Interruptroutine sichern
     old_vect11= getvect(irq11); //alte Interruptroutine sichern
     setvect(irq10, interruptmotor);//neue Interruptroutine einfgen
     setvect(irq11, interruptlight);//neue Interruptroutine einfgen
     outportb(slavemask, inportb(slavemask) & 0xF3);
				 //Interrupts freigeben (Bit 2 und 3 lschen)
     enable();                   //Interrupts allgemein anschalten
  };

/*Ŀ
    Interruptroutine Motor                                                
  */
  void interrupt interruptmotor(void)
  {
    disable();                  //Interrupts abschalten whrend der Routine
    irqstatus = 10;             //Interrupt 10 ist eingetroffen
    outportb(master, eoi);      //End-Of-Interrupt and Interr.-Controller 1
    outportb(slave, eoi);       //End-Of-Interrupt and Interr.-Controller 2
    enable();                   //Interrupts allgemein anschalten
  };

/*Ŀ
    Interruptroutine Lichtschranke                                        
  */
  void interrupt interruptlight(void)
  {
    disable();                  //Interrupts abschalten whrend der Routine
    irqstatus = 11;             //Interrupt 11 ist eingetroffen
    outportb(master, eoi);      //End-Of-Interrupt and Interr.-Controller 1
    outportb(slave, eoi);       //End-Of-Interrupt and Interr.-Controller 2
    enable();                   //Interrupts allgemein anschalten
  };

/*Ŀ
    Interrupts sperren und alte Interruptroutine wieder eintragen         
  */
  void resetinter (void)
  {
    outportb(slavemask, inportb(slavemask) | 0xC);
				//Interrupts sperren (Bit 2 und 3 setzen)
    setvect(irq10, old_vect10); //alte Interruptroutine einfgen
    setvect(irq11, old_vect11); //alte Interruptroutine einfgen
  };

/*Ŀ
    alle Ports auf Eingang                                                
  */
  void resetports (void)
  {

  };


/*Ŀ
   LoadRam (ldt Sinusfunktionswerte in RAM 1 und Cosinusfkt in RAM 2)    
  */
  void loadram(void)
  {

  };


/*͸
    Hauptprogramm                                                         
  ;*/
  int main(void)
  {


    return 0;
  };
